/*
 *  DVDSubEdit and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVDSubEdit and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "VTSClean.h"
#include <stdio.h>
#include <objbase.h>
#include "Cell.h"
#include "IFOUtils.h"
#include "Utils.h"
#include <io.h>

typedef unsigned char uchar;

#define VERS "FixVTS -- 1.603"

// This is the beginning of a stuffing pack.
uchar StuffingPackHead[20] = {0x00,0x00,0x01,0xBA,0x44,0x00,0x04,0x00,0x04,0x01,0x01,0x89,0xC3,0xF8,0x00,0x00,0x01,0xBE,0x07,0xEC};

#define CHAR_SIZE 256
// This stuff is global and accessed by all the classes, for messages and errors.
// The string where messages are stored.
char Message[CHAR_SIZE];

FILE* CVTSClean::m_Outlog = 0;

// Our default function to handle regular and error messages.
void DefaultLogMessage(char* AMessage, int Langage)
{
	if(Langage == FRENCH)
		return;
	if(AMessage[strlen(AMessage)-1]=='\n') AMessage[strlen(AMessage)-1] = '\0';
	printf(AMessage);
	strcpy(Message,AMessage);
	if(CVTSClean::m_Outlog)
	{
		fprintf(CVTSClean::m_Outlog,"%s\n",AMessage);
		fflush(CVTSClean::m_Outlog);
	}
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}
// Global pointers that classes use to send regular and error messages.
void (*pMessageFunction)(char* AMessage, int Langage) = DefaultLogMessage;
void (*pErrorFunction)(char* AMessage, int Langage) = DefaultLogMessage;
int	 (*FrameSelectFunction)(int m_VOBID, int m_CELLID, int StartLBA, int EndLBA, int Default, 
							int NumLBA, int NormallySkipped) = 0;
#define MAX_NUM_NAVPACKS 512*1024
#define MAX_NUM_CELLS 1024*16

#ifdef _DEBUG
#define MYASSERT(A,B) {assert(A);}
#else
#define MYASSERT(A,B) {if(!(A)) pErrorFunction(B,0);}
#endif


CVTSClean::CVTSClean()
{
	m_Running = 0;
	m_AbortNow = 0;
	m_NumInputLBA = 1;
	m_InPackIndex = 0;
	m_Phase = Processing_Idle;
	strcpy(m_FileName,"");
	strcpy(m_Version,VERS);
	m_RemoveUnrefed = 0;
	m_RemoveBlank = 0;
	m_CumulLastNavPack = 0;
	m_Pause = 0;
	m_InPlaceProcess = 0;
	m_MenuDomain = 0;
	m_NumFwdPtrFixed = 0;
	m_NumPacksRemoved = 0;
	m_CumulNumInputLBA = 0;
	m_RemapVCID = 1;
	m_RemapVOBID = 1;
	m_ResetILVUPtrs = 0;
	m_WaitForFile = 0;
	m_PutBackupAbove = 0;
	m_FixPGCLBAPtrs = 1;

	// This is our stuffing pack...
	memset(m_StuffingPack,0xFF,SECTOR_SIZE);
	memcpy(m_StuffingPack,StuffingPackHead,20);
	strcpy(m_LogPath,"");

	m_VCIDMap = new VCIDMapItem[MAX_NUM_CELLS];
	m_NextCID = new uchar[MAX_NUM_CELLS];
}
	
CVTSClean::~CVTSClean()
{
	delete[] m_VCIDMap;
	delete[] m_NextCID;
}

void CVTSClean::Reset()
{
	m_OutPackIndex = 0;
	m_CumulOutPackIndex = 0;
	m_CumulInPackIndex = 0;
	m_LastNavPack = 0;
	m_CumulLastNavPack = 0;
	m_LastNavPackFwdPtr = -1;
	m_VOBID = 0;
	m_CELLID = 0;
	m_NumFwdPtrFixed = 0;
	m_NumPacksRemoved = 0;
	m_NumPacksRemapped = 0;
	m_NumILVUPtrReset = 0;
	m_SkipThisCell = false;

	m_NewCELLID = 0;
	m_VCIDMapLength = 0;
	// All CID should start at 1
	memset(m_NextCID,1,MAX_NUM_CELLS);
	m_NextVID = 1;
}


int CVTSClean::OpenFile(const char* FileName)
{
	// Only accept VOB files...
	if(strcmpi(GetExt(FileName),".VOB") != 0 || !PathExists(FileName))
		return -1;
	strcpy(m_FileName,FileName);
	if(strcmpi(GetFname(m_FileName),"video_ts") == 0 || m_FileName[strlen(m_FileName)-5] == '0')
		m_MenuDomain = 1;
	else
	{
		// Make sure we're opening the first file of the domain.
		m_MenuDomain = 0;
		m_FileName[strlen(m_FileName)-5] = '1';
	}
	VideoFormat = FORMAT_UNKNOWN;
	return 0;
}

int CVTSClean::CountInputLBAs()
{
	int Count = GetFileSizekB(m_FileName)/2;
	char FileName[1024];
	strcpy(FileName,m_FileName);
	while(FindNextVOB(FileName) != 0)
		Count += GetFileSizekB(FileName)/2;
	return Count;
}

// This creates the path for the backup version of FileName.
char* CVTSClean::GetBackupName(char* FileName)
{
	char DirPath[1024];
	strcpy(DirPath,GetDirPath(FileName));
	if(m_PutBackupAbove)
	{
		// Go from end and find first occurence of \ or /
		for(int i=strlen(DirPath)-3;i>1;i--)
			if(DirPath[i] == '\\' || DirPath[i] == '/')
			{
				DirPath[i+1]=0;
				break;
			}
	}
	if(!PathExists(CatStr(DirPath,"FixVTSBackup")))
		_mkdir(CatStr(DirPath,"FixVTSBackup"));
	return CatStr(DirPath,"FixVTSBackup\\",GetFnameExt(FileName));
}

void CVTSClean::RestoreThisFile(char* FileName)
{
	// Both original and backup exit. Remove original and move backup.
	if(PathExists(GetBackupName(FileName)) && PathExists(FileName))
		remove(FileName);
	// Now move backup file back.
	rename(GetBackupName(FileName),FileName);
}

void CVTSClean::RestoreBackup()
{
#if 0
	// Moves everything back into the original directory
	struct _finddata_t FileInfo;
	int FindIt = _findfirst( CatPath(GetDirPath(GetBackupName(m_FileName)),"*"), &FileInfo );

	char BackupPath[1024];
	char DirPath[1024];
	strcpy(BackupPath,GetDirPath(GetBackupName(m_FileName)));
	strcpy(DirPath,GetDirPath(m_FileName));

	if(strcmp(FileInfo.name,".") != 0 && strcmp(FileInfo.name,"..") != 0)
	{
		if(PathExists(CatPath(DirPath,FileInfo.name)))
			remove(CatPath(DirPath,FileInfo.name));
		rename(CatStr(BackupPath,FileInfo.name),CatStr(DirPath,FileInfo.name));
	}
	while(_findnext(FindIt, &FileInfo ) == 0)
	{
		if(strcmp(FileInfo.name,".") != 0 && strcmp(FileInfo.name,"..") != 0)
		{
			if(PathExists(CatPath(DirPath,FileInfo.name)))
				remove(CatPath(DirPath,FileInfo.name));
			rename(CatStr(BackupPath,FileInfo.name),CatStr(DirPath,FileInfo.name));
		}
	}
#else
	// Here, I'm only restoring the currently process file..
	// First, restore IFO.
	if(!m_MenuDomain)
		m_FileName[strlen(m_FileName)-5] = '1';
	RestoreThisFile(GetIFO(m_FileName));
	RestoreThisFile(GetBUP(m_FileName));
	// Then restore video_ts.ifo
	RestoreThisFile(CatStr(GetDirPath(m_FileName),"VIDEO_TS.IFO"));
	RestoreThisFile(CatStr(GetDirPath(m_FileName),"VIDEO_TS.BUP"));
	// Then restore all the files.
	RestoreThisFile(m_FileName);
	while(FindNextVOB(m_FileName,false))
		RestoreThisFile(m_FileName);

#endif	

	MESSAGE1("Backup files restored");
}

void CVTSClean::DeleteBackup()
{
	// Empty backup directory.
	EmptyDir(GetDirPath(GetBackupName(m_FileName)));
	_rmdir(GetDirPath(GetBackupName(m_FileName)));
}

bool CVTSClean::FindNextVOB(char* FileName, bool CheckEmpty)
{
	if(m_MenuDomain)
		return 0;
	while(FileName[strlen(FileName)-5] != '9')
	{
		FileName[strlen(FileName)-5]++;
		if(!CheckEmpty || GetFileSizekB(FileName) > 0)
			return 1;
	}
	return 0;
}

char* FindPreviousVOB(char* FileName)
{
	static char Temp[1024];
	strcpy(Temp,FileName);
	Temp[strlen(Temp)-5]--;
	return Temp;
}

int CVTSClean::ReadOpenThisFile(char* FileName, bool FromBackup)
{
	if(FromBackup)
	{
		// Opening with backup. If the backup exists, we delete it, then copy original to
		// backup.
		if(PathExists(GetBackupName(m_FileName)))
			// Backup exists... Delete it!
			remove(GetBackupName(m_FileName));
		if(rename(m_FileName,GetBackupName(m_FileName)) != 0)
		{
			ERROR2("Couldn't backup %s",m_FileName);
			m_InFileFD = 0;
			return 1;
		}
		m_NumInputLBA = GetFileSizekB(GetBackupName(m_FileName)) / 2;
		m_InFileFD = fopen(GetBackupName(m_FileName),"rb");
		if(!m_InFileFD)
			ERROR2("Couldn't open %s for read/write",GetBackupName(m_FileName));
		return m_InFileFD ? 0 : 1;
	}
	else
	{
		// In place processing, simpler...
		if(m_WaitForFile)
		{
			// In this mode, FixVTS waits until the 
			int First = 1;
			do
			{
				Sleep(100);
				m_InFileFD = fopen(m_FileName,"r+b");
				if(First == 1 && m_InFileFD == 0)
					MESSAGE2("Waiting for file %s to be ready for processing",GetFname(m_FileName));
				First = 0;
			}
			while(m_InFileFD == NULL && m_AbortNow != TRUE);
			MESSAGE2("File %s now ready for processing",GetFname(m_FileName));
		}
		else
		{
			m_InFileFD = fopen(m_FileName,"r+b");
			if(!m_InFileFD)
				ERROR2("Couldn't open %s for read/write",m_FileName);

		}
		m_NumInputLBA = GetFileSizekB(m_FileName) / 2;
		return m_InFileFD ? 0 : 1;
	}
}

int CVTSClean::WriteOpenThisFile(char* FileName)
{
	m_OutFileFD = fopen(m_FileName,"w+b");
	if(!m_OutFileFD)
		ERROR2("Couldn't open %s for writing",m_FileName);
	return m_OutFileFD ? 0 : 1;
}

CIFOLoader* CVTSClean::OpenIFO(char* FileName)
{
	CIFOLoader *IFO;
	if(!m_InPlaceProcess)
	{
		// Load IFO
		IFO = new CIFOLoader(GetIFO(m_FileName));
		// Write into backup dir...
		if(IFO->m_LoadedOK)
		{
			IFO->WriteIFO(GetBackupName(GetIFO(m_FileName)));
			IFO->CloseIFO();
		}
	}
	else
	{
		// In-place processing. Don't backup the IFOs.
		IFO = new CIFOLoader(GetIFO(m_FileName));
	}
	return IFO;
}

void CVTSClean::UpdateLastNavPack(int CumulNavPackLBA, int ForwardPointer)
{
//	MYASSERT(m_MenuDomain == false,"Can't update last navpack in menu domain!");
	int NavPackLBA;
	char FileName[1024];
	strcpy(FileName,m_FileName);
	if(m_MenuDomain)
	{
		int ThisSize = GetFileSizekB(FileName)/2;
		assert(ThisSize > CumulNavPackLBA);
		NavPackLBA = CumulNavPackLBA;
	}
	else
	{
		// First, find in which vob the navpack is that needs to be updated
		int i;
		int CumulSizeLBA = 0;
		for(i=1;i<10;i++)
		{
			// This isn't very good programming!
			FileName[strlen(FileName)-5]='0'+(char)i;
			// Get size and see if this LBA is in this file...
			int ThisSize = GetFileSizekB(FileName)/2;
			if(CumulSizeLBA + ThisSize > CumulNavPackLBA)
				break;
			CumulSizeLBA += ThisSize;
		}
		if(i==10)
		{
			MYASSERT(0,"Couln't find the right file to update last navpack");
			return;
		}

		// NavPackLBA will be the LBA in this file.
		NavPackLBA = CumulNavPackLBA - CumulSizeLBA;
	}

	FILE* fd = fopen(FileName,"r+b");
	MYASSERT(fd != 0,"Couldn't open VOB to update last navpack");
	if(!fd)
		return;
	fseek(fd,NavPackLBA*SECTOR_SIZE,SEEK_SET);
	fread(m_Prev,SECTOR_SIZE,1,fd);
	PutInt(m_Prev,0x040f,ForwardPointer);
	fseek(fd,NavPackLBA*SECTOR_SIZE,SEEK_SET);
	fwrite(m_Prev,SECTOR_SIZE,1,fd);
	fclose(fd);
}

int CVTSClean::OnNewCell(CIFOLoader* IFO)
{
	// A new cell was found.
	MESSAGE5("Found VCID %d/%d at LBA %d (%d out)",m_VOBID,m_CELLID,m_InPackIndex, m_OutPackIndex);
	if(m_CellInfos[m_NumCells].m_VOBID != 0 && !m_SkipThisCell)
	{
		// We finished reading the previous cell. If it's not to be skipped, let's complete its info
		m_CellInfos[m_NumCells].m_EndLBA = m_CumulOutPackIndex-1;
		// and renumber the cell.
		ushort NewVOBID;
		uchar NewCELLID;
		GetMapping(m_CellInfos[m_NumCells].m_VOBID, m_CellInfos[m_NumCells].m_CELLID, NewVOBID, NewCELLID);
		m_CellInfos[m_NumCells].m_VOBID = NewVOBID;
		m_CellInfos[m_NumCells].m_CELLID = NewCELLID;
		m_NumCells++;
//MESSAGE1("POINT N1");
		MYASSERT(m_NumCells < MAX_NUM_CELLS,"Too many cells!");
	}
	// Now gather info for new cell.
	m_CellInfos[m_NumCells].m_VOBID = m_VOBID;
	m_CellInfos[m_NumCells].m_CELLID = m_CELLID;
	m_CellInfos[m_NumCells].m_StartLBA = m_CumulOutPackIndex;
	// Let's see if this cell is referenced in any PGC. If it isn't, then remove it.
	m_SkipThisCell = false;
//MESSAGE1("POINT N2");
	if(IFO->m_LoadedOK && IFO->IsReferenced(m_VOBID,m_CELLID,m_MenuDomain) == 0)
	{
		// This cell is unreferenced...
		MESSAGE3("VCID %d/%d is unreferenced...",m_VOBID,m_CELLID);
		if(m_RemoveUnrefed)
			m_SkipThisCell = true;
	}
	if(!m_SkipThisCell)
	{
//MESSAGE1("POINT N3");
		// If we're not skipping this cell, let's add it to our VCID map.
		if(!AlreadyInVCIDMap(m_VOBID,m_CELLID))
		{
			uint NewVCID = AddToVCIDMap(m_VOBID,m_CELLID);
			if(NewVCID != m_CELLID+256u*m_VOBID)
				MESSAGE5("Remapping %d/%d to %d/%d",m_VOBID,m_CELLID,NewVCID/256,NewVCID%256);
		}
	}
	return 0;
}

bool CVTSClean::ResetILVUPtrs()
{
	bool DidReset = 0;
	// SML_PBI ILVU flag:
	DidReset |= (GetShort(m_Data,0x427) != 0);
	memset(m_Data+0x427,0,12);

	// Angle information for non-seamless play
	memset(m_Data+0x69,0,36);

	// Angle information for seamless jumps
	memset(m_Data+0x4bb,0,54);

	if(DidReset)
		m_NumILVUPtrReset++;
	return DidReset;
}

int CVTSClean::ProcessPack(CIFOLoader* IFO)
{
	CPack ThePack;

	// Identify pack and retrieve video format if needed (we'll need that for IFO processing).
	ThePack.FindPackType(m_Data,m_VOBID,m_CELLID);
	if(VideoFormat == FORMAT_UNKNOWN && ThePack.m_PackType == PackTypeVideo)
		ThePack.FindFrameType(m_Data,0);

	// Did we reach a new cell?
	if(m_CellInfos[m_NumCells].m_VOBID != m_VOBID || m_CellInfos[m_NumCells].m_CELLID != m_CELLID)
		OnNewCell(IFO);
//MESSAGE1("POINT 1");
	// If we're skipping this cell, we have nothing to do. But let's still remember the input
	// output navpack mapping. This will be necessary for the TMAT table.
	if(m_SkipThisCell)
	{
		if(ThePack.m_PackType == PackTypeNavPack)
		{
			// This is a navpack. Let's keep track of where we are.
			m_InNavPacks[m_NavPackCount] = m_CumulInPackIndex;
			m_OutNavPacks[m_NavPackCount++] = m_CumulOutPackIndex;
			MYASSERT(m_NavPackCount < MAX_NUM_NAVPACKS,"Too many navpacks!");
//MESSAGE1("POINT 1a");
		}
		m_NumPacksRemoved++;
		return 0;
	}
//MESSAGE1("POINT 2");

	// These will store the two this->lba pointers in navpacks.
	uint ThisLBA1 = 0;
	uint ThisLBA2 = 0;
	bool NeedRemap = 0;
	bool ILVUReset = 0;
	ushort NewVOBID = 0;
	uchar NewCELLID = 0;
	if(ThePack.m_PackType == PackTypeNavPack)
	{
		// This is a navpack. Let's keep track of where we are. This will be used to correct
		// the PGC info in the IFOs.
		m_InNavPacks[m_NavPackCount] = m_CumulInPackIndex;
		m_OutNavPacks[m_NavPackCount++] = m_CumulOutPackIndex;
		MYASSERT(m_NavPackCount < MAX_NUM_NAVPACKS,"Too many navpacks!");
//MESSAGE1("POINT 3");

		// Find previous navpack, and adjust its forward pointer if necessary.
		if(m_CumulOutPackIndex > m_OutPackIndex && m_LastNavPack >= m_OutPackIndex)
		{
			// This will happen at the boundary between 2 consecutive vob files.
			// I need to go back to the previous file, and update the last navpack pointer.
			if(m_LastNavPackFwdPtr != (m_CumulOutPackIndex - m_CumulLastNavPack -1))
			{
				MESSAGE2("Adjusting Fwd pointer in previous VOB at LBA %d",m_LastNavPack);
				MYASSERT(m_MenuDomain == false,"Trying to adjust previous file, but in menu domain!");
				UpdateLastNavPack(m_CumulLastNavPack, m_CumulOutPackIndex - m_CumulLastNavPack -1);
				m_NumFwdPtrFixed++;
			}
		}
		else if(m_LastNavPackFwdPtr != (m_OutPackIndex - m_LastNavPack -1))
		{
//MESSAGE1("POINT 3a");
			// Seek back to previous navpack, read sector, adjust pointers, seek back and write.
			fseek(m_OutFileFD,m_LastNavPack*SECTOR_SIZE,SEEK_SET);
			fread(m_Prev,SECTOR_SIZE,1,m_OutFileFD);
			PutInt(m_Prev,0x040f,m_OutPackIndex - m_LastNavPack -1);
			fseek(m_OutFileFD,m_LastNavPack*SECTOR_SIZE,SEEK_SET);
			fwrite(m_Prev,SECTOR_SIZE,1,m_OutFileFD);
			// Now seek back to where we were. If not in place, we're seeking the output pointer.
			// But if in place, we need to seek to the point we were at...
			if(!m_InPlaceProcess)
				fseek(m_OutFileFD,m_OutPackIndex*SECTOR_SIZE,SEEK_SET);
			else
				fseek(m_OutFileFD,(m_InPackIndex+1)*SECTOR_SIZE,SEEK_SET);
			MESSAGE2("Adjusting Fwd pointer at LBA %d",m_LastNavPack);
			m_NumFwdPtrFixed++;
		}
//MESSAGE1("POINT 4");

		// Keep track of the navpack LBA, and the two this->lba pointers.
		m_LastNavPack = m_OutPackIndex;
		m_CumulLastNavPack = m_CumulOutPackIndex;
		ThisLBA1 = GetInt(m_Data,0x2d);
		ThisLBA2 = GetInt(m_Data,0x40b);
		// Also adjust these two pointers in the current pack. May not be needed if this->lba is correct.
		PutInt(m_Data,0x2d,m_CumulOutPackIndex);
		PutInt(m_Data,0x40b,m_CumulOutPackIndex);
		// And keep a copy of the forward pointer.
		m_LastNavPackFwdPtr = GetInt(m_Data,0x040f);
		// Now let's see if we need to remap the VCID:
		NeedRemap = GetMapping(m_VOBID, m_CELLID, NewVOBID, NewCELLID);
		if(NeedRemap)
			m_NumPacksRemapped++;
		// Write new VCID to buffer.
		PutShort(m_Data,0x41f,NewVOBID);
		m_Data[0x422] = NewCELLID;
		// Now do the angle stuff if needed.
//MESSAGE1("POINT 4a");
		if(m_ResetILVUPtrs)
			ILVUReset = ResetILVUPtrs();
	}
//MESSAGE1("POINT 5");

	// Now's the time to write the pack.
	if(m_InPlaceProcess)
	{
		// If in place: write pack only if pointers are wrong, or if inlba != outlba. 
		if(m_OutPackIndex != m_InPackIndex ||				// write if we're "moving" packs.
			(ThePack.m_PackType == PackTypeNavPack && (		// or it's a nav pack and...
				(int)ThisLBA1 != m_CumulOutPackIndex ||		// the pointers are wrong...
				(int)ThisLBA2 != m_CumulOutPackIndex) ||	// or we need to remap the VCID.
				NeedRemap ||								// or we reset the angle pointers.
				ILVUReset))
		{
			int ret = fseek(m_OutFileFD,m_OutPackIndex*SECTOR_SIZE,SEEK_SET);
			ret = fwrite(m_Data,1,SECTOR_SIZE,m_OutFileFD);
			// Seek back in pointer (same as out) so next read is in right spot.
			ret = fseek(m_InFileFD,(m_InPackIndex+1)*SECTOR_SIZE,SEEK_SET);
//MESSAGE1("POINT 6");

		}
//MESSAGE1("POINT 7");

		// Uncomment this to remove stuffing packs .
//		if(GetInt(m_Data,0x0e) == 0x000001be) ThePack.m_PackType = PackTypeBlank;
		// Increase output index only if pack isn't skipped.
		if(!m_RemoveBlank || ThePack.m_PackType != PackTypeBlank)
		{
			// We don't want to remove blanks, or this pack isn't a blank pack.

			if(ThePack.m_PackType == PackTypeBlank)
			{
				// We don't want to remove blanks and this pack is a blank pack:
				// Replace blank pack with stuffing pack.
				int ret = fseek(m_OutFileFD,m_OutPackIndex*SECTOR_SIZE,SEEK_SET);
				ret = fwrite(m_StuffingPack,1,SECTOR_SIZE,m_OutFileFD);
				// Seek back in pointer (same as out) so next read is in right spot.
				ret = fseek(m_InFileFD,(m_InPackIndex+1)*SECTOR_SIZE,SEEK_SET);
				static int warned = 0;
				if(warned++ < 30)
					MESSAGE2("Replacing blank pack at LBA %d with stuffing pack",m_InPackIndex);
			}
			// Increase output pointers.
			m_OutPackIndex++;
			m_CumulOutPackIndex++;
		}
		else
		{
			// We want to remove blanks, and this pack is a blank pack.
			static int warned = 0;
			if(warned++ < 30)
				MESSAGE2("Skipping blank pack at LBA %d",m_InPackIndex);
			m_NumPacksRemoved++;
		}
	}
	else
	{
//MESSAGE1("POINT 8");
		// If not in place: write pack only if pack isn't skipped.
		if(!m_RemoveBlank || ThePack.m_PackType != PackTypeBlank)
		{
			// We don't want to remove blanks, or this pack isn't a blank pack.
			int ret = fseek(m_OutFileFD,m_OutPackIndex*SECTOR_SIZE,SEEK_SET);
			if(ThePack.m_PackType == PackTypeBlank)
			{
				// We don't want to remove blanks and this pack is a blank pack:
				// Replace blank pack with stuffing pack.
				MESSAGE2("Replacing blank pack at LBA %d with stuffing pack",m_InPackIndex);
				ret = fwrite(m_StuffingPack,1,SECTOR_SIZE,m_OutFileFD);
			}
			else
			{
//MESSAGE1("POINT 9");
				// This pack is not a blank pack, write it.
				ret = fwrite(m_Data,1,SECTOR_SIZE,m_OutFileFD);
			}
			// Increase output pointers.
			m_OutPackIndex++;
			m_CumulOutPackIndex++;
		}
		else
		{
			// We want to remove blanks, and this pack is a blank pack.
			MESSAGE2("Skipping blank pack at LBA %d",m_InPackIndex);
			m_NumPacksRemoved++;
		}
	}
//MESSAGE1("POINT 10");
	return 0;
}

void TruncateFile(char* FileName, uint FileSizeLBA)
{
	// This reopens FileName and truncates it to the size indicated in FileSizeLBA.
	HANDLE Temp = CreateFile(FileName,GENERIC_READ|GENERIC_WRITE, 
		FILE_SHARE_READ,NULL,OPEN_ALWAYS,0,NULL);
	SetFilePointer(Temp,FileSizeLBA*SECTOR_SIZE,0,FILE_BEGIN);
	SetEndOfFile(Temp);
	CloseHandle(Temp);
}

void CVTSClean::RemoveEmptyVobs()
{
//MESSAGE1("REMOVING EMPTY VOBS");
	if(m_MenuDomain)
	{
		// It's possible for a menu to become completely empty. This will remove the vob file.
		if(IsEmpty(m_FileName))
		{
			if(PathExists(m_FileName))
				MESSAGE2("%s is empty... removing.",GetFname(m_FileName));
			remove(m_FileName);
		}
		return;
	}

	// Get VTS number.
	int ret, VTS,Index;
	if((ret=sscanf(GetFname(m_FileName),"%*[^_]_%d_%d",&VTS,&Index)) != 2 || VTS > 99)
	{
//MESSAGE4("REMOVING EMPTY VOBS Ret %d, VTS %d Index %d",ret,VTS,Index);
		MYASSERT(0,"Problem while checking file sizes");
		return;
	}
//MESSAGE4("REMOVING EMPTY VOBS Ret %d, VTS %d Index %d",ret,VTS,Index);
	char* DirName = GetDirPath(m_FileName);
	char InFileName[1024];
	char OutFileName[1024];
	int InIndex;
	int OutIndex;
	for(InIndex=OutIndex=1;InIndex<10;InIndex++)
	{
		sprintf(InFileName,"%sVTS_%02d_%d.VOB",DirName,VTS,InIndex);
		sprintf(OutFileName,"%sVTS_%02d_%d.VOB",DirName,VTS,OutIndex);
//MESSAGE2("Testing %s ",InFileName);
//MESSAGE2("IsEmpty %d",IsEmpty(InFileName));
		if(!IsEmpty(InFileName))
		{
			// Rename file, even if it's the same name...
			if(strcmpi(InFileName,OutFileName) != 0)
				MESSAGE3("Renaming %s to %s",GetFname(InFileName),GetFname(OutFileName));
			rename(InFileName,OutFileName);
			OutIndex++;
		}
		else
		{
			// Remove this file!
			if(PathExists(InFileName))
				MESSAGE2("%s is empty... removing.",GetFname(InFileName));
			remove(InFileName);
		}
	}
}


void CVTSClean::Process()
{
	m_Pause = 0;
	m_Outlog = fopen(m_LogPath,"a");
	if(m_Outlog) fprintf(m_Outlog,"%s\n",m_Version);
	// Input file name stuff.
	if(!m_MenuDomain)
		m_FileName[strlen(m_FileName)-5] = '1';

	m_CumulNumInputLBA = CountInputLBAs();

	// Open input and output files.
	if(!m_InPlaceProcess)
	{
		if(ReadOpenThisFile(m_FileName,true) != 0)
			return;
		if(WriteOpenThisFile(m_FileName) != 0)
			return;
	}
	else
	{
		if(ReadOpenThisFile(m_FileName,false) != 0)
			return;
		m_OutFileFD = m_InFileFD;
	}

	// At this point, if you remove unrefed, you also remove blanks etc...
	m_RemoveBlank = m_RemoveUnrefed;
	CIFOLoader	*IFO = OpenIFO(m_FileName);
	if(!IFO->m_LoadedOK)
	{
		MESSAGE1("Warning: Couldn't load IFO file..");
	}

	// Allocate array of info on cells.
	m_CellInfos = new VTSM_C_ADTItem[MAX_NUM_CELLS];
	memset(m_CellInfos,0,MAX_NUM_CELLS*sizeof(VTSM_C_ADTItem));
	m_NumCells = 0;
	// Arrays of navpack lba infos.
	m_InNavPacks = new int[MAX_NUM_NAVPACKS];
	m_OutNavPacks = new int[MAX_NUM_NAVPACKS];
	m_NavPackCount = 0;

	// Now, go through the packs.
	int ret;
	Reset();

	// Main loop over packs...
	m_Phase = Processing_VOB;
	MESSAGE3("Processing %s -- %s",m_FileName,m_InPlaceProcess?"in-place processing":"from backup");
	MESSAGE2("Remove unrefed is %s",m_RemoveUnrefed?"ON":"OFF");
	MESSAGE2("Reset ILVU Ptrs is %s",m_ResetILVUPtrs?"ON":"OFF");
	for(m_InPackIndex=0;;m_InPackIndex++,m_CumulInPackIndex++)
	{
		// Simple-minded pause mechanism.
		while(m_Pause && !m_AbortNow)
			Sleep(10);
		if(m_AbortNow) break;
		// Read next pack.
		ret = (fread(m_Data,1,SECTOR_SIZE,m_InFileFD) == SECTOR_SIZE);
		if(ret != 1)
		{
			// Couldn't read pack.
			// Find out if there's more files to read...
			fclose(m_InFileFD);
			fclose(m_OutFileFD);
			// Truncate file if in-place processing.
			if(m_InPlaceProcess && m_InPackIndex != m_OutPackIndex)
				TruncateFile(m_FileName,m_OutPackIndex);
			// Open next vob file.
			if(FindNextVOB(m_FileName) == 1)
			{
				MESSAGE2("Processing %s",m_FileName);
				// Open input and output files.
				if(!m_InPlaceProcess)
				{
					if(ReadOpenThisFile(m_FileName,true) != 0)
						return;
					if(WriteOpenThisFile(m_FileName) != 0)
						return;
				}
				else
				{
					if(ReadOpenThisFile(m_FileName,false) != 0)
						return;
					m_OutFileFD = m_InFileFD;
				}
				// Now read pack.
				ret = (fread(m_Data,1,SECTOR_SIZE,m_InFileFD) == SECTOR_SIZE);
				MYASSERT(ret == 1, "Couldn't read m_Data");
				// Reset these guys.
				m_OutPackIndex = 0;
				m_InPackIndex = 0;
			}
			else
			{
				// This was to update last navpack forward pointer.
				if(m_LastNavPackFwdPtr != (m_CumulOutPackIndex - m_CumulLastNavPack -1))
				{
					UpdateLastNavPack(m_CumulLastNavPack,m_CumulOutPackIndex - m_CumulLastNavPack -1);
				}
				break;
			}
		}
		
		// Process this pack.
		ProcessPack(IFO);
	}

	MESSAGE5("Done with VOB files. %d navpack pointers adjusted, %d packs removed, %d navpack remapped, %d navpacks ILVU flags reset",
		m_NumFwdPtrFixed,m_NumPacksRemoved, m_NumPacksRemapped, m_NumILVUPtrReset);

	// We're done. process IFO files.
	fclose(m_InFileFD);
	fclose(m_OutFileFD);

	if(!m_AbortNow)
	{
		RemoveEmptyVobs();
		ProcessIFOs(IFO);
	}
	else
		m_Phase = Processing_Aborted;

	MESSAGE1("Done processing");
	MESSAGE1("Cleaning up");
	// Restore file name...
	if(!m_MenuDomain)
		m_FileName[strlen(m_FileName)-5] = '1';
	delete[] m_InNavPacks;
	delete[] m_OutNavPacks;
	delete[] m_CellInfos;
	delete IFO;
	MESSAGE1("Done");
	if(m_Outlog) fclose(m_Outlog);
}


int CVTSClean::ProcessIFOs(CIFOLoader* IFO)
{
	if(!IFO->m_LoadedOK)
	{
		MESSAGE2("Couldn't process IFO file %s",GetIFO(m_FileName));
		return -1;
	}
	MESSAGE2("Processing %s",GetIFO(m_FileName));
	m_Phase = Processing_IFO;
//	Sleep(400);
	// Remap the last cell.
	ushort NewVOBID;
	uchar NewCELLID;
	if(!m_SkipThisCell && m_CumulOutPackIndex > 0)
	{
		GetMapping(m_CellInfos[m_NumCells].m_VOBID, m_CellInfos[m_NumCells].m_CELLID, NewVOBID, NewCELLID);
		m_CellInfos[m_NumCells].m_VOBID = NewVOBID;
		m_CellInfos[m_NumCells].m_CELLID = NewCELLID;
		// Finish gathering info for the last cell.
		m_CellInfos[m_NumCells++].m_EndLBA = m_CumulOutPackIndex-1;
	}
	// I also need to do this! Otherwise the PGCInfos of the last cell will be wrong. I'm faking the
	// presence of a navpack at the very end of the input and output vobs.
	m_InNavPacks[m_NavPackCount] = m_CumulInPackIndex;
	m_OutNavPacks[m_NavPackCount++] = m_CumulOutPackIndex;

	// Now, remap the VCIDs.
	IFO->RenumberVCID(m_VCIDMap,m_VCIDMapLength,m_MenuDomain);
	// Now reconstruct the entire VTS_C_ADT table.
	MESSAGE1("Now cleaning VTS_C_ADT"); 
	IFO->RebuildVTS_C_ADT(m_CellInfos,m_NumCells,m_MenuDomain);
	// This will go through all the PGCs, and update the Cell pointers according to the
	// navpack input->output map. It will also reconstruct the VOBUADMAP table.
	MESSAGE1("Now updating navpack info"); 
	IFO->UpdateNavPackInfo(m_InNavPacks,m_OutNavPacks,m_NavPackCount,m_MenuDomain);
	// Then, this will correct the PGC cell pointers to make sure they fall within the cells they're
	// supposed to.
	if(m_FixPGCLBAPtrs)
	{
		MESSAGE1("Checking PGC LBA pointers"); 
		IFO->UpdateCellInfo(m_CellInfos,m_NumCells,m_MenuDomain);
	}

	if(!m_MenuDomain)
	{
		MESSAGE1("Now updating navpack info"); 
		IFO->UpdateTMAPTI(m_InNavPacks,m_OutNavPacks,m_NavPackCount);
	}

	MESSAGE1("Now updating PGC playback times"); 
	// Now update the PGC playback time in case it's faulty.
	if(!m_MenuDomain)
	{
		int NumPGC = IFO->GetNumPGC(0);
		for(int i=1;i<=NumPGC;i++)
			IFO->UpdatePGCPlayBackTime(0,(ushort)i,true);
	}
	else
	{
		int NumLU = IFO->GetNumLU();
		for(int j=1;j<=NumLU;j++)
		{
			int NumPGC = IFO->GetNumPGC(j);
			for(int i=1;i<=NumPGC;i++)
			IFO->UpdatePGCPlayBackTime(j,(ushort)i,true);
		}
	}

	// Update local VTS sectors.
	MESSAGE1("Now updating VTS sectors"); 
	IFO->UpdateVTSSectors();
	m_Phase = Writing_IFO;
//	Sleep(400);
	MESSAGE1("Now writing IFO file"); 
	// Write back IFO. The backup will have been done in the OpenIFO() call.
	IFO->WriteIFO(GetIFO(m_FileName));
	// Now update VIDEO_TS.IFO->..
	char Video_ts[512];
	strcpy(Video_ts,CatStr(GetDirPath(m_FileName),"VIDEO_TS.IFO"));
	if(PathExists(Video_ts))
	{
		IFO->LoadIFO(Video_ts);
		// Copy backup if not in place, and if we didn't already backup!
		if(!m_InPlaceProcess && strcmpi(GetFname(m_FileName),"video_ts") != 0)
			IFO->WriteIFO(GetBackupName(Video_ts));
		MESSAGE1("Now updating video_ts.ifo vts sectors"); 
		// Correct number of VTSs if necessary.
		IFO->UpdateVTSSectors(true);
		IFO->CloseIFO();
		// Now write it back.
		MESSAGE1("Now writing video_ts.ifo"); 
		IFO->WriteIFO(Video_ts);
	}
	m_Phase = Processing_Done;
	return 0;
}

void CVTSClean::UpdateIFOForEmptyVob()
{
	// Empty vobs, let's update the IFO just the same if it exists...
	if(PathExists(GetIFO(m_FileName)))
	{
		// IFO exists, load it.
		CIFOLoader *IFO = OpenIFO(m_FileName);
		if(IFO->m_LoadedOK)
		{
			// IFO loaded OK. Let's zero out the C_ADT and VOBUM_ADMAP tables.
			MESSAGE2("%s does not exists, updating IFO...",m_FileName);
			m_SkipThisCell = 0;
			m_NavPackCount = m_CumulOutPackIndex = m_CumulInPackIndex = 0;
			m_InNavPacks = new int[10];
			m_OutNavPacks = new int[10];
			m_VCIDMapLength = 0;
			m_NumCells = 0;
			ProcessIFOs(IFO);
			delete[] m_InNavPacks;
			delete[] m_OutNavPacks;
		}
		delete IFO;
	}
}

void CVTSClean::ProcessFullDVD() 
{
	int i;

	char DirPath[1024];
	strcpy(DirPath,GetDirPath(m_FileName));
	
	MESSAGE1("PROCESSING FULL DVD"); 
	for(i=1;i<=99;i++)
	{
		// Menu
		if(m_WaitForFile && i == 1)
		{
			// If in "wait for file" mode, wait for first titleset to be ripped.
			bool First = 1;
			sprintf(m_FileName,"%sVTS_%02d_1.VOB",DirPath,i);
			MESSAGE1("Waiting for VTS 1 to be ready for processing");
			while(!PathExists(m_FileName) && m_AbortNow == false)
			{
				Sleep(100);
			}
		}
		sprintf(m_FileName,"%sVTS_%02d_0.VOB",DirPath,i);
		// I really want the base class function here! No user interaction!
		if(CVTSClean::OpenFile(m_FileName) == 0)
			Process();
		else
		{
			// No vob file, let's see if there's an IFO.
			m_MenuDomain = 1;
			UpdateIFOForEmptyVob();
		}
		while(m_Pause && !m_AbortNow)
			Sleep(10);
		if(m_AbortNow) return;
		// Title
		sprintf(m_FileName,"%sVTS_%02d_1.VOB",DirPath,i);
		if(CVTSClean::OpenFile(m_FileName) == 0)
			Process();
		else
		{
			// No vob file, let's see if there's an IFO.
			m_MenuDomain = 0;
			UpdateIFOForEmptyVob();
		}
		while(m_Pause && !m_AbortNow)
			Sleep(10);
	}

	// Now process video_ts.vob
	sprintf(m_FileName,"%sVIDEO_TS.VOB",DirPath);
	if(CVTSClean::OpenFile(m_FileName) == 0)
		Process();
	else
	{
		// No vob file, let's see if there's an IFO.
		m_MenuDomain = 1;
		UpdateIFOForEmptyVob();
	}
	MESSAGE1("End of FULL DVD process"); 
}

bool CVTSClean::AlreadyInVCIDMap(ushort VOBID, uchar CELLID)
{
	int i;
	// Go through VCIDmap, and see whether we're already encountered the VCID.
	for(i=0;i<m_VCIDMapLength;i++)
	{
		if(VOBID == m_VCIDMap[i].InVOBID && CELLID == m_VCIDMap[i].InCELLID)
			return true;
	}
	return false;

}

// This version also updates the VOBID!
uint CVTSClean::AddToVCIDMap(ushort VOBID, uchar CELLID)
{
	// This adds the VCID to our map, assigning the next Cell ID for the given VOBID.
	MYASSERT(m_VCIDMapLength < MAX_NUM_CELLS, "VCIDMap too small!");
	MYASSERT(AlreadyInVCIDMap(VOBID,CELLID) == false,"Adding VCID already in map!");
	m_VCIDMap[m_VCIDMapLength].InVOBID = VOBID;
	m_VCIDMap[m_VCIDMapLength].InCELLID = CELLID;
	
	if(m_RemapVCID && m_RemapVOBID)
	{
		// Remapping VOB and CELL IDs. Check if the VOBID was previously seen.
		int i;
		for(i=0;i<m_VCIDMapLength;i++)
			if(VOBID == m_VCIDMap[i].InVOBID)
				break;
		if(i==m_VCIDMapLength)
			// This is a new VOBID. Let's remap it to the next one available.
			VOBID = m_VCIDMap[m_VCIDMapLength].OutVOBID = m_NextVID++;
		else
			// We already saw that one. Reusing VOBID.
			VOBID = m_VCIDMap[m_VCIDMapLength].OutVOBID = m_VCIDMap[i].OutVOBID;
	}
	else
	{
		// No remapping. Out is the same as in.
		m_VCIDMap[m_VCIDMapLength].OutVOBID = VOBID;
	}

	if(m_RemapVCID)
	{
		// Out CID is taken from m_NextCID:
		m_VCIDMap[m_VCIDMapLength].OutCELLID = m_NextCID[VOBID];
		m_VCIDMapLength++;
		// Increment m_NextCID[VOBID].
		m_NextCID[VOBID] += 1;
		return m_NextCID[VOBID]-1 + 256*VOBID;
	}
	else
	{
		// No remapping. Same output as input...
		m_VCIDMap[m_VCIDMapLength].OutCELLID = CELLID;
		m_VCIDMapLength++;
		return CELLID + 256*VOBID;
	}
}


bool CVTSClean::GetMapping(ushort VOBID, uchar CELLID, ushort &NewVOBID, uchar &NewCELLID)
{
	int i;
	// Go through our map and find the desired VCID...
	for(i=0;i<m_VCIDMapLength;i++)
	{
		if(VOBID == m_VCIDMap[i].InVOBID && CELLID == m_VCIDMap[i].InCELLID)
		{
			// Found it.
			NewCELLID = m_VCIDMap[i].OutCELLID;
			NewVOBID = m_VCIDMap[i].OutVOBID;
			// Return true if the new cell id isn't the same as the original!
			return (NewCELLID != CELLID || NewVOBID != VOBID);
		}
	}
	// Should never end up here!
	if(m_VCIDMapLength)
		MYASSERT(0, "GetMapping failed!");
	return false;
}

void CVTSClean::EraseLog() // Erases the log file.
{
}
