#ifndef INCLUDES_H
#define INCLUDES_H

#include <stdio.h>
#include <string.h>
#include <assert.h>

// This include file is used by all the other ones in this project.

#define SECTOR_SIZE 2048

// A few typedefs.
typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;

// Types of Packs we'll deal with. Blank packs can happen in arccos protected DVDs.
typedef enum
{
	PackTypeNone,
	PackTypeNavPack,
	PackTypeVideo,
	PackTypeAudio,
	PackTypeSupPic,
	PackTypePadding,
	PackTypeBlank,
} PackType;

// Types of frames for video packs.
typedef enum
{
	FrameTypeNone,
	FrameTypeI,
	FrameTypeP,
	FrameTypeB,
} FrameType;

// This is only used for I-frames.
typedef enum
{
	IFrameTypeNone,
	IFrameTypeStart,		// The first sector of an I-Frame.
	IFrameTypeMiddle,		// A middle sector of an I-Frame.
	IFrameTypeEnd,			// The last sector of an I-Frame.
} IFrameType;

// Typedef for video formats.
typedef enum {
	FORMAT_UNKNOWN,
	FORMAT_PAL,
	FORMAT_NTSC
} VFormat;

// Typedef used to handle the SRC format.
typedef struct
{
	ushort	m_SRCFrac;	// Fractional part of the SRC, remainder of the complete SRC divided by 300
	uint	m_SRCInt;	// Integer part of the SRC.
	uchar	m_SRCTop;	// Just one bit. The full SRC is encoded on 33 bits!!!
} CompositeSRC;

#ifdef WIN32
#include <string.h>
#include <direct.h>

// Endian definitions for various start codes. Rather than convert to endian agnostic values,
// it's fast to look at the raw value, but what you compare to has to be changed accordingly.
// PC values. JEAN: WHAT ABOUT LINUX?! Little endian values.
#define PICT_START_CODE 0x00010000
#define MPEG_START_CODE 0xb3010000
#define END_OF_SEQ_CODE 0xb7010000
#define PADD_START_CODE 0xbe010000
#define GOP__START_CODE 0xb8010000
#define USER_START_CODE 0xb2010000
#define CCAP_START_CODE 0xf8014343	// Indicates closed captions. Found after GOP, and after a USER_START_CODE
// See http://www.geocities.com/mcpoodle43/SCC_TOOLS/DOCS/SCC_FORMAT.HTML

#else
// Mac values. BigEndian values.
#define PICT_START_CODE 0x00000100
#define MPEG_START_CODE 0x000001b3
#define END_OF_SEQ_CODE 0x000001b7
#define PADD_START_CODE 0x000001be
#define GOP__START_CODE 0x000001b8
#define USER_START_CODE 0x000001b2
#define CCAP_START_CODE 0x434301f8
// Proper includes for posix:
#include <sys/stat.h>
#include <unistd.h>
#define _mkdir(A) mkdir(A,0755);
#endif


// Functions to read and write values into buffers, taking care of little-endian issues.
extern void PutInt(uchar* Buffer, int Offset, uint Value);
extern void PutShort(uchar* Buffer, int Offset, ushort Value);
extern uint GetInt(uchar* Buffer, int Offset);
extern ushort GetShort(uchar* Buffer, int Offset);
// Caution, no check for wrap-around, for example, adding or subtracting too much!
void AddShort(uchar* Buffer, int Offset, short AddValue);
void AddInt(uchar* Buffer, int Offset, int AddValue);
// These ones convert a uint or a ushort on the fly.
extern uint FixEndian(uint Input);
extern ushort FixEndian(ushort Input);

// This is a global declared in VOBClasses.cpp
extern VFormat VideoFormat;	// NTSC etc.

// This is a pointer to a function that can be used to select which I-frame to use. 
// It should return the LBA of the chosen I-frame, -1 to kill the cell and -2 to keep it animated 
// (unprocessed).
// Parameters are:
// The VOB ID, the Cell ID, the start and end sectors of the cell, the default I-frame sector,
// the cell size in sectors, and whether the cell was going to be killed (no buttons).
extern int	 (*FrameSelectFunction)(int VobId, int CellId, int StartLBA, int EndLBA, int Default, 
									int SizeLBA, int NormallySkipped);

// Error and messages. Eventually you could redefine these as you please. 
// Message, pMessageFunction and pErrorFunction are defined in MenuShrinkEngine.cpp. 
// But you can redefine pMessageFunction and pErrorFunction to point to your own functions.
extern char Message[];
extern void (*pMessageFunction)(char* AMessage, int Langage);
extern void (*pErrorFunction)(char* AMessage, int Langage);

#define ENGLISH		0
#define FRENCH		1

// This is a bit low-tech, 
#define ERROR1(A) {sprintf(Message,A); pErrorFunction(Message, ENGLISH);}
#define ERROR2(A,B) {sprintf(Message,A,B); pErrorFunction(Message, ENGLISH);}
#define ERROR3(A,B,C) {sprintf(Message,A,B,C); pErrorFunction(Message, ENGLISH);}
#define ERROR4(A,B,C,D) {sprintf(Message,A,B,C,D); pErrorFunction(Message, ENGLISH);}

#define MESSAGE1(A) {sprintf(Message,A); pMessageFunction(Message, ENGLISH);}
#define MESSAGE2(A,B) {sprintf(Message,A,B); pMessageFunction(Message, ENGLISH);}
#define MESSAGE3(A,B,C) {sprintf(Message,A,B,C); pMessageFunction(Message, ENGLISH);}
#define MESSAGE4(A,B,C,D) {sprintf(Message,A,B,C,D); pMessageFunction(Message, ENGLISH);}
#define MESSAGE5(A,B,C,D,E) {sprintf(Message,A,B,C,D,E); pMessageFunction(Message, ENGLISH);}
#define MESSAGE6(A,B,C,D,E,F) {sprintf(Message,A,B,C,D,E,F); pMessageFunction(Message, ENGLISH);}

#define ERRORF1(A) {sprintf(Message,A); pErrorFunction(Message, FRENCH);}
#define ERRORF2(A,B) {sprintf(Message,A,B); pErrorFunction(Message, FRENCH);}
#define ERRORF3(A,B,C) {sprintf(Message,A,B,C); pErrorFunction(Message, FRENCH);}
#define ERRORF4(A,B,C,D) {sprintf(Message,A,B,C,D); pErrorFunction(Message, FRENCH);}

#define MESSAGEF1(A) {sprintf(Message,A); pMessageFunction(Message, FRENCH);}
#define MESSAGEF2(A,B) {sprintf(Message,A,B); pMessageFunction(Message, FRENCH);}
#define MESSAGEF3(A,B,C) {sprintf(Message,A,B,C); pMessageFunction(Message, FRENCH);}
#define MESSAGEF4(A,B,C,D) {sprintf(Message,A,B,C,D); pMessageFunction(Message, FRENCH);}
#define MESSAGEF5(A,B,C,D,E) {sprintf(Message,A,B,C,D,E); pMessageFunction(Message, FRENCH);}

#ifdef SILENT
#undef MESSAGE2
#define MESSAGE2(A,B) {;}
#undef MESSAGE3
#define MESSAGE3(A,B,C) {;} 
#undef MESSAGE4
#define MESSAGE4(A,B,C,D) {;}
#undef MESSAGE5
#define MESSAGE5(A,B,C,D,E) {;}
#undef MESSAGE6
#define MESSAGE6(A,B,C,D,E,F) {;}
#endif
#endif
